"""Net Functions.

The following functions give you access to interact with http services.
"""

from __future__ import print_function

__all__ = [
    "getExternalIpAddress",
    "getHostName",
    "getIpAddress",
    "getRemoteServers",
    "httpClient",
    "httpDelete",
    "httpGet",
    "httpPost",
    "httpPut",
    "openURL",
    "sendEmail",
]

import socket

from typing import Any, Callable, Dict, List, Optional

from com.inductiveautomation.ignition.common.script.builtin.http import JythonHttpClient
from java.lang import IllegalArgumentException, String


def getExternalIpAddress():
    # type: () -> String
    """Returns the client's IP address, as it is detected by the
    Gateway.

    This means that this call will communicate with the Gateway, and the
    Gateway will tell the client what IP address its incoming traffic is
    coming from. If you have a client behind a Network Address
    Translation (NAT) router, then this address will be the Wide Area
    Network (WAN) address of the router instead of the Local Area
    Network (LAN) address of the client, which is what you'd get with
    system.net.getIpAddress.

    Returns:
        A text representation of the Client's IP address, as detected by
        the Gateway.
    """
    return "52.52.32.221"


def getHostName():
    # type: () -> String
    """Returns the host name of the computer that the script was ran
    on.

    When run in the Gateway scope, returns the Gateway hostname. When
    run in the Client scope, returns the Client hostname. On Windows,
    this is typically the "computer name". For example, might return
    EAST_WING_WORKSTATION or bobs-laptop.

    Returns:
        The hostname of the local machine.
    """
    return socket.gethostname()


def getIpAddress():
    # type: () -> String
    """Returns the IP address of the computer that the script was ran
    on.

    Returns:
        Returns the IP address of the local machine, as it sees it.
    """
    return socket.gethostbyname(str(getHostName()))


def getRemoteServers(runningOnly=True):
    # type: (Optional[bool]) -> List[String]
    """This function returns a List of Gateway Network servers that are
    visible from the local Gateway.

    Args:
        runningOnly: If set to True, only servers on the Gateway Network
            that are running will be returned. Servers that have lost
            contact with the Gateway Network will be filtered out.
            Optional.

    Returns:
        A List of strings representing Gateway Network Server IDs.
    """
    print(runningOnly)
    return []


def httpClient(
    timeout=60000,  # type: int
    bypass_cert_validation=False,  # type: bool
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    proxy=None,  # type: Optional[String]
    cookie_policy="ACCEPT_ORIGINAL_SERVER",  # type: String
    redirect_policy="NORMAL",  # type: String
    version="HTTP_2",  # type: String
    customizer=None,  # type: Optional[Callable[..., Any]]
):
    # type: (...) -> JythonHttpClient
    """Provides a general use object that can be used to send and
    receive HTTP requests.

    The object created by this function is a wrapper around Java's
    HttpClient class. Usage requires creating a JythonHttpClient object
    with a call to system.net.httpClient, then calling a method (such as
    get(), post()) on the JythonHttpClient to actually issue a request.

    Args:
        timeout: A value, in milliseconds, to set the client's connect
            timeout setting to. Defaults to 60000. Optional.
        bypass_cert_validation: A boolean indicating whether the client
            should attempt to validate the certificates of remote
            servers, if connecting via HTTPS/SSL. Defaults to False.
            Optional.
        username: A string indicating the username to use for
            authentication if the remote server requests authentication;
            specifically, by responding with a WWW-Authenticate or
            Proxy-Authenticate header. Only supports Basic
            authentication. If username is specified but not password,
            an empty string will be used for the password in the Basic
            Authentication response. Defaults to None. Optional.
        password: A string indicating the password to use for
            authentication. Defaults to None. Optional.
        proxy: The address of a proxy server, which will be used for
            HTTP and HTTPS traffic. If a port is not specified as part
            of that address, it will be assumed from the protocol in the
            URL, i.e. 80/443. Defaults to None. Optional.
        cookie_policy: A string representing this client's cookie
            policy. Accepts values "ACCEPT_ALL", "ACCEPT_NONE", and
            "ACCEPT_ORIGINAL_SERVER". Defaults to
            "ACCEPT_ORIGINAL_SERVER". Optional.
        redirect_policy: A string representing this client's redirect
            policy. Acceptable values are listed below. Defaults to
            "Normal". Optional.
        version: A string specifying either HTTP_2 or HTTP_1_1 for the
            HTTP protocol. When omitted, the previous default of HTTP_2
            is implied. Optional.
        customizer: A reference to a function. This function will be
            called with one argument (an instance of
            HttpClient.Builder). The function should operate on that
            builder instance, which allows for customization of the
            created HTTP client. Defaults to None. Optional.

    Returns:
        An object wrapped around an instance of Java's HttpClient class.
        The httpClient object has methods that can be called to execute
        HTTP requests against a server.
    """
    print(
        timeout,
        bypass_cert_validation,
        username,
        password,
        proxy,
        cookie_policy,
        redirect_policy,
        version,
        customizer,
    )
    return JythonHttpClient()


def httpDelete(
    url,  # type: String
    contentType=None,  # type: Optional[String]
    connectTimeout=10000,  # type: int
    readTimeout=60000,  # type: int
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    headerValues=None,  # type: Optional[Dict[String, String]]
    bypassCertValidation=True,  # type: bool
):
    # type: (...) -> String
    """Performs an HTTP DELETE to the given URL.

    Args:
        url: The URL to send the request to.
        contentType: The MIME type used in the HTTP 'Content-type'
            header. Optional.
        connectTimeout: The timeout for connecting to the URL in
            milliseconds. Default is 10,000. Optional.
        readTimeout: The read timeout for the operation in
            milliseconds. Default is 60,000. Optional.
        username: If specified, the call will attempt to authenticate
            with basic HTTP authentication. Optional.
        password: The password used for basic HTTP authentication, if
            the username parameter is also present. Optional.
        headerValues: A dictionary of name/value pairs that will be set
            in the HTTP header. Optional.
        bypassCertValidation: If the target address in an HTTPS address,
            and this parameter is TRUE, the system will bypass all SSL
            certificate validation. This is not recommended, though is
            sometimes necessary for self-signed certificates. Optional.

    Returns:
        The content returned for the DELETE operation.
    """
    print(
        url,
        contentType,
        connectTimeout,
        readTimeout,
        username,
        password,
        headerValues,
        bypassCertValidation,
    )
    return "DELETE"


def httpGet(
    url,  # type: String
    connectTimeout=10000,  # type: int
    readTimeout=60000,  # type: int
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    headerValues=None,  # type: Optional[Dict[String, String]]
    bypassCertValidation=None,  # type: Optional[bool]
    useCaches=True,  # type: bool
    throwOnError=True,  # type: bool
):
    # type: (...) -> String
    """Retrieves the document at the given URL using the HTTP GET
    protocol.

    The document is returned as a string. For example, if you use the
    URL of a website, you'll get the same thing you'd get by going to
    that website in a browser and using the browser's "View Source"
    function.

    Args:
        url: The URL to retrieve.
        connectTimeout: The timeout for connecting to the URL. In
            milliseconds. Default is 10,000. Optional.
        readTimeout: The read timeout for the get operation. In
            milliseconds. Default is 60,000. Optional.
        username: If specified, the call will attempt to authenticate
            with basic HTTP authentication. Optional.
        password: The password used for basic HTTP authentication, if
            the username parameter is also present. Optional.
        headerValues: A dictionary of name/value pairs that will be set
            in the HTTP header. Optional.
        bypassCertValidation: If the target address is an HTTPS address,
            and this parameter is True, the system will bypass all SSL
            certificate validation. This is not recommended, though is
            sometimes necessary for self-signed certificates. Optional.
        useCaches: Will cache the information returned by the httpGet
            call. If using this for something that constantly updates
            like an rss feed, it would be better to set this to False.
            Default is True. Optional.
        throwOnError: Set to False if you wish to get the error body
            rather than a Python exception if the GET request returns an
            error code (non-200 responsive). Default is True. Optional.

    Returns:
        The content found at the given URL.
    """
    print(
        url,
        connectTimeout,
        readTimeout,
        username,
        password,
        headerValues,
        bypassCertValidation,
        useCaches,
        throwOnError,
    )
    return ""


def httpPost(url, *args, **kwargs):
    # type: (String, *Any, **Any) -> String
    """Retrieves the document at the given URL using the HTTP POST
    protocol.

    If a parameter dictionary argument is specified, the entries in the
    dictionary will encoded in "application/x-www-form-urlencoded"
    format, and then posted. You can post arbitrary data as well, but
    you'll need to specify the MIME type. The document is then returned
    as a string.

    Args:
        url: The URL to post to.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        The content returned for the POST operation.
    """
    print(url, args, kwargs)
    return ""


def httpPut(url, *args, **kwargs):
    # type: (String, *Any, **Any) -> String
    """Performs an HTTP PUT to the given URL.

    Encodes the given dictionary of parameters using
    "applications/x-www-form-urlencoded" format.

    Args:
        url: The URL to put to.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        The content returned by the PUT operation.
    """
    print(url, args, kwargs)
    return ""


def openURL(url, useApplet=False):
    # type: (String, Optional[bool]) -> None
    """Opens the given URL or URI scheme outside of the currently
    running Client in whatever application the host operating system
    deems appropriate.

    Args:
        url: The URL to open in a web browser.
        useApplet: If set to True, and the client is running as an
            Applet, then the browser instance that launched the applet
            will be used to open the URL. Optional.
    """
    print(url, useApplet)


def _sendEmail(
    smtpSettings,  # type: Optional[String]
    fromAddr,  # type: String
    subject=None,  # type: Optional[String]
    body=None,  # type: Optional[String]
    html=False,  # type: bool
    to=None,  # type: Optional[List[String]]
    attachmentNames=None,  # type: Optional[List[object]]
    attachmentData=None,  # type: Optional[List[object]]
    timeout=300000,  # type: int
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    priority="3",  # type: String
    cc=None,  # type: Optional[List[String]]
    bcc=None,  # type: Optional[List[String]]
    retries=0,  # type: int
    replyTo=None,  # type: Optional[List[String]]
):
    # type: (...) -> None
    _to = [] if to is None else to
    _cc = [] if cc is None else cc
    _bcc = [] if bcc is None else bcc
    recipients = _to + _cc + _bcc
    if smtpSettings and fromAddr and len(recipients) > 0:
        print(
            smtpSettings,
            fromAddr,
            subject,
            body,
            html,
            to,
            attachmentNames,
            attachmentData,
            timeout,
            username,
            password,
            priority,
            cc,
            bcc,
            retries,
            replyTo,
        )
    else:
        raise IllegalArgumentException(
            "Cannot send email without SMTP host, from address and recipient list."
        )


def sendEmail(
    smtp=None,  # type: Optional[String]
    fromAddr="",  # type: String
    subject=None,  # type: Optional[String]
    body=None,  # type: Optional[String]
    html=False,  # type: bool
    to=None,  # type: Optional[List[String]]
    attachmentNames=None,  # type: Optional[List[object]]
    attachmentData=None,  # type: Optional[List[object]]
    timeout=300000,  # type: int
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    priority="3",  # type: String
    smtpProfile=None,  # type: Optional[String]
    cc=None,  # type: Optional[List[String]]
    bcc=None,  # type: Optional[List[String]]
    retries=0,  # type: int
    replyTo=None,  # type: Optional[List[String]]
):
    # type: (...) -> None
    """Sends an email through the given SMTP server.

    Note that this email is relayed first through the Gateway - the
    client host machine doesn't need network access to the SMTP server.

    Args:
        smtp: The address of an SMTP server to send the email through,
            like "mail.example.com". A port can be specified, like
            "mail.example.com:25". SSL can also be forced, like
            "mail.example.com:25:tls".
        fromAddr: An email address to have the email come from.
        subject: The subject line for the email. Optional.
        body: The body text of the email. Optional.
        html: A flag indicating whether or not to send the email as an
            HTML email. Will auto-detect if omitted. Optional.
        to: A list of email addresses to send to.
        attachmentNames: A list of attachment names. Attachment names
            must have the correct extension for the file type or an
            error will occur. Optional.
        attachmentData: A list of attachment data, in binary format.
        timeout: A timeout for the email, specified in milliseconds.
            Defaults to 5 minutes (60,000*5). Optional.
        username: If specified, will be used to authenticate with the
            SMTP host. Optional.
        password: If specified, will be used to authenticate with the
            SMTP host. Optional.
        priority: Priority for the message, from "1" to "5", with "1"
            being highest priority. Defaults to "3" (normal) priority.
            Optional.
        smtpProfile: If specified, the named SMTP profile defined
            in the Gateway will be used. If this keyword is present, the
            smtp, username, and password keywords will be ignored.
            Optional.
        cc: A list of email addresses to carbon copy. Only available if
            a smtpProfile is used. Optional.
        bcc: A list of email addresses to blind carbon copy. Only
            available if a smtpProfile is used. Optional.
        retries: The number of additional times to retry sending on
            failure. Defaults to 0. Only available if a smtpProfile is
            used. Optional.
        replyTo: An optional list of addresses to have the recipients
            reply to. If omitted, this defaults to the from address.
            Optional.
    """
    if not smtpProfile:
        _sendEmail(
            smtp,
            fromAddr,
            subject,
            body,
            html,
            to,
            attachmentNames,
            attachmentData,
            timeout,
            username,
            password,
            priority,
            retries=retries,
            replyTo=replyTo,
        )
    else:
        _sendEmail(
            smtpProfile,
            fromAddr,
            subject,
            body,
            html,
            to,
            attachmentNames,
            attachmentData,
            timeout,
            priority=priority,
            cc=cc,
            bcc=bcc,
            retries=retries,
            replyTo=replyTo,
        )



from java.awt import Component, Graphics
from java.lang import Object


class ComponentUI(Object):
    def __init__(self):
        # type: () -> None
        pass

    def contains(self, c, x, y):
        # type: (Component, int, int) -> bool
        pass

    def paint(self, g, c):
        # type: (Graphics, Component) -> None
        pass

    def update(self, g, c):
        # type: (Graphics, Component) -> None
        pass


class DesktopPaneUI(ComponentUI):
    def __init__(self):
        # type: () -> None
        pass



"""Roster Functions.

Functions that provide roster manipulation, including adding and remove
users from a roster.
"""

from __future__ import print_function

__all__ = [
    "addUsers",
    "createRoster",
    "deleteRoster",
    "getRoster",
    "getRosterNames",
    "getRosters",
    "removeUsers",
]

from typing import Dict, List

from com.inductiveautomation.ignition.alarming.common.rosters import RosterModel
from com.inductiveautomation.ignition.common.user import PyUser
from java.lang import String


def addUsers(rosterName, users):
    # type: (String, List[PyUser]) -> None
    """Adds a list of users to an existing roster.

    Users are always appended to the end of the roster.

    Args:
        rosterName: The name of the roster to modify.
        users: A list of User objects that will be added to the end of
            the roster. User objects can be created with the
            system.user.getUser and system.user.addUser functions. These
            users must exist before being added to the roster.
    """
    print(rosterName)
    for user in users:
        print(user.Username)


def createRoster(name, description):
    # type: (String, String) -> None
    """Creates a roster with the given name and description, if it does
    not already exist.

    This function was designed to run in the Gateway and in Perspective
    sessions. If creating rosters from Vision clients, use
    system.alarm.createRoster instead.

    Args:
        name: The name of the roster to create.
        description: The description for the roster. May be None, but
            the parameter is mandatory.
    """
    print(name, description)


def deleteRoster(rosterName):
    # type: (String) -> None
    """Deletes a roster with the given name.

    Args:
        rosterName: The name of the roster to delete.
    """
    print(rosterName)


def getRoster(name):
    """Returns the roster corresponding to the given name.

    Args:
        name: The name of the roster to get the RosterModel object from.

    Returns:
        A RosterModel Object. Call getUsers() to access a list of User
        objects in the RosterModel.
    """
    print(name)
    return RosterModel()


def getRosterNames():
    # type: () -> List[String]
    """Returns a list of roster names.

    Returns:
        A list of Strings representing the roster names
    """
    return ["On-Call"]


def getRosters():
    # type: () -> Dict[String, List[String]]
    """Returns a dictionary of rosters, where the key is the name of the
    roster, and the value is an array list of string user names.

    This function was designed to run in the Gateway and in Perspective
    sessions. If creating rosters from Vision clients, use
    system.alarm.getRosters instead.

    Returns:
        A dictionary that maps roster names to a list of usernames in
        the roster. The list of usernames may be empty if no users have
        been added to the roster.
    """
    return {}


def removeUsers(rosterName, users):
    # type: (String, List[PyUser]) -> None
    """Removes one or more users from an existing roster.

    Args:
        rosterName: The name of the roster to modify.
        users: A list of user objects that will be added to the end of
            the roster. User objects can be created with the
            system.user.getUser and system.user.addUser functions.
    """
    print(rosterName)
    for user in users:
        print(user.Username)



__all__ = ["DataQuality", "DataType"]

from typing import List, Optional

from com.inductiveautomation.ignition.common.i18n import LocalizedString
from com.inductiveautomation.ignition.common.model.values import Quality, QualityCode
from java.lang import Enum, Object, String
from java.util import Locale


class DataQuality(Object):
    @staticmethod
    def fromIntValue(value):
        # type: (int) -> DataQuality
        pass

    def getDescription(self):
        # type: () -> LocalizedString
        pass

    def getIntValue(self):
        # type: () -> int
        pass

    def getLevel(self):
        # type: () -> Quality.Level
        pass

    def getName(self):
        # type: () -> str
        pass

    def getQualityCode(self):
        # type: () -> QualityCode
        pass

    @staticmethod
    def getQualityFor(value):
        # type: (int) -> DataQuality
        pass

    def isDataUsed(self):
        # type: () -> bool
        pass

    def isGood(self):
        # type: () -> bool
        pass

    def isGoodData(self):
        # type: () -> bool
        pass

    def isOpcBadData(self):
        # type: () -> bool
        pass

    def toString(self, locale=None):
        # type: (Optional[Locale]) -> String
        pass

    @staticmethod
    def valueOf(name):
        # type: (str) -> DataQuality
        pass

    @staticmethod
    def values():
        # type: () -> List[DataQuality]
        pass

    @staticmethod
    def worstOf(q1, q2):
        # type: (DataQuality, DataQuality) -> DataQuality
        pass

    @staticmethod
    def worstOfAll(*args):
        # type: (DataQuality) -> DataQuality
        pass


class DataType(Enum):
    def fromIntValue(self, val):
        pass

    def getComponentDataType(self):
        pass

    def getIntValue(self):
        pass

    def getJavaType(self):
        pass

    def getTypeClass(self):
        pass

    def getTypeForClass(self, clazz):
        pass

    def getTypeForValue(self, val):
        pass

    def isArray(self):
        pass

    def isFloatingPoint(self):
        pass

    def isNumeric(self):
        pass

    def legacyDataType(self):
        pass

    def values(self):
        pass



__all__ = [
    "Channel",
    "FileChannel",
    "SeekableByteChannel",
    "SelectableChannel",
    "SelectionKey",
    "Selector",
    "SocketChannel",
]

from typing import Any, Optional, Set, TypeVar, Union

from java.lang import AutoCloseable, Object
from java.nio import ByteBuffer, MappedByteBuffer
from java.util.function import Consumer

T = TypeVar("T")


class Channel(AutoCloseable):
    def close(self):
        # type: () -> None
        pass

    def isOpen(self):
        # type: () -> bool
        pass


class ReadableByteChannel(Channel):
    def read(self, dst):
        # type: (ByteBuffer) -> int
        pass


class WriteableByteChannel(Channel):
    def write(self, src):
        # type: (ByteBuffer) -> int
        pass


class FileChannel(Object, Channel):
    def force(self, metaData):
        # type: (bool) -> None
        pass

    def lock(self):
        # type: () -> FileLock
        pass

    def map(self, mode, position, size):
        # type: (FileChannel.MapMode, long, long) -> MappedByteBuffer
        raise NotImplementedError

    @staticmethod
    def open(*args):
        # type: (*Any) -> FileChannel
        pass

    def position(self, newPosition=None):
        # type: (Optional[long]) -> Union[FileChannel, long]
        pass

    def read(self, *args):
        # type: (Any) -> Union[int, long]
        pass

    def size(self):
        # type: () -> long
        pass

    def transferFrom(self, src, position, count):
        # type: (ReadableByteChannel, long, long) -> long
        pass

    def transferTo(self, position, count, target):
        # type: (long, long, WriteableByteChannel) -> long
        pass

    def truncate(self, size):
        # type: (long) -> FileChannel
        pass

    def tryLock(
        self,
        position=None,  # type: Optional[long]
        size=None,  # type: Optional[long]
        shared=None,  # type: Optional[bool]
    ):
        # type: (...) -> FileLock
        pass

    def write(self, *args):
        # type: (Any) -> Union[int, long]
        pass

    class MapMode(Object):
        PRIVATE = None  # type: FileChannel.MapMode
        READ_ONLY = None  # type: FileChannel.MapMode
        READ_WRITE = None  # type: FileChannel.MapMode


class FileLock(Object, AutoCloseable):
    def acquiredBy(self):
        # type: () -> Channel
        pass

    def channel(self):
        # type: () -> FileChannel
        pass

    def close(self):
        # type: () -> None
        pass

    def isShared(self):
        # type: () -> bool
        pass

    def isValid(self):
        # type: () -> bool
        raise NotImplementedError

    def overlaps(self, position, size):
        # type: (long, long) -> bool
        pass

    def position(self):
        # type: () -> long
        pass

    def release(self):
        # type: () -> None
        raise NotImplementedError

    def size(self):
        # type: () -> long
        pass


class SeekableByteChannel(ReadableByteChannel, WriteableByteChannel):
    def position(self, newPosition=None):
        # type: (Optional[long]) -> Union[long, SeekableByteChannel]
        pass

    def size(self):
        # type: () -> long
        pass

    def truncate(self, size):
        # type: (long) -> SeekableByteChannel
        pass


class SelectableChannel(Object, AutoCloseable):
    def close(self):
        # type: () -> None
        pass

    def blockingLock(self):
        # type: () -> Object
        pass

    def configureBlocking(self):
        # type: () -> SelectableChannel
        pass

    def isBlocking(self):
        # type: () -> bool
        pass

    def isRegistered(self):
        # type: () -> bool
        pass

    def keyFor(self, sel):
        # type: (Selector) -> SelectionKey
        pass

    def provider(self):
        # type: () -> Object
        pass

    def register(self, sel, ops, att=None):
        # type: (Selector, int, Optional[Object]) -> SelectionKey
        pass

    def validOps(self):
        # type: () -> int
        pass


class SelectionKey(Object):
    OP_ACCEPT = None  # type: int
    OP_CONNECT = None  # type: int
    OP_READ = None  # type: int
    OP_WRITE = None  # type: int

    def attach(self, ob):
        # type: (Object) -> Object
        pass

    def attachment(self):
        # type: () -> Object
        pass

    def cancel(self):
        # type: () -> None
        pass

    def channel(self):
        # type: () -> SelectableChannel
        pass

    def interestOps(self, ops=None):
        # type: (Optional[int]) -> Union[int, SelectionKey]
        pass

    def interestOpsAnd(self, ops):
        # type: (int) -> int
        pass

    def interestOpsOr(self, ops):
        # type: (int) -> int
        pass

    def isAcceptable(self):
        # type: () -> bool
        pass

    def isConnectable(self):
        # type: () -> bool
        pass

    def isReadable(self):
        # type: () -> bool
        pass

    def isValid(self):
        # type: () -> bool
        pass

    def isWriteable(self):
        # type: () -> bool
        pass

    def readyOps(self):
        # type: () -> int
        pass

    def selector(self):
        # type: () -> Selector
        pass


class Selector(Object, AutoCloseable):
    def close(self):
        # type: () -> None
        pass

    def isOpen(self):
        # type: () -> bool
        pass

    def keys(self):
        # type: () -> Set[SelectionKey]
        pass

    @staticmethod
    def open():
        # type: () -> Selector
        pass

    def provider(self):
        # type: () -> Object
        pass

    def select(self, *args):
        # type: (Any) -> int
        pass

    def selectedKeys(self):
        # type: () -> Set[SelectionKey]
        pass

    def selectNow(self, action=None):
        # type: (Optional[Consumer]) -> int
        pass

    def wakeup(self):
        # type: () -> Selector
        pass


class SocketChannel(SelectableChannel):
    def bind(self, local):
        # type: (Any) -> SocketChannel
        pass

    def connect(self, remote):
        # type: (Any) -> bool
        pass

    def finishConnect(self):
        # type: () -> bool
        pass

    def getLocalAddress(self):
        # type: () -> Any
        pass

    def getRemoteAddress(self):
        # type: () -> Any
        pass

    def isConnected(self):
        # type: () -> bool
        pass

    def isConnectionPending(self):
        # type: () -> bool
        pass

    def open(self, remote=None):
        # type: (Optional[Any]) -> SocketChannel
        pass

    def read(self, *args):
        # type: (Any) -> Union[int, long]
        pass

    def setOption(self, name, value):
        # type: (Any, T) -> SocketChannel
        pass

    def shutdownInput(self):
        # type: () -> SocketChannel
        pass

    def shutdownOutput(self):
        # type: () -> SocketChannel
        pass

    def socket(self):
        # type: () -> Any
        pass

    def write(self, *args):
        # type: (Any) -> Union[int, long]
        pass



from typing import Any

from java.lang import Object


class Binding(object):
    def update(self, *args):
        # type: (*Any) -> None
        raise NotImplementedError


class DefaultBindableModel(Object):
    def bind(self, toBind):
        # type: (Binding) -> None
        pass

    def modelUpdated(self, *args):
        # type: (*Any) -> None
        pass

    def unbind(self, toUnbind):
        # type: (Binding) -> None
        pass

    def unbindAll(self):
        # type: () -> None
        pass

    def update(self):
        # type: () -> None
        pass



"""OPC - UA Functions.

The following functions allow you to interact directly with an OPC UA
Server.
"""

from __future__ import print_function

__all__ = ["addConnection", "callMethod", "removeConnection"]

from typing import Any, Dict, List, Tuple

from java.lang import String


def addConnection(
    name,  # type: String
    description,  # type: String
    discoveryUrl,  # type: String
    endpointUrl,  # type: String
    securityPolicy,  # type: String
    securityMode,  # type: String
    settings,  # type: Dict[String, Any]
):
    # type: (...) -> None
    """Adds a new OPC UA connection.

    Args:
        name: Name to assign to the new connection.
        description: Description assigned to the new OPC UA connection.
        discoveryUrl: Endpoint URL to use for discovery services.
        endpointUrl: Endpoint URL to use for session services.
        securityPolicy: The name of the SecurityPolicy to use.
        securityMode: The name of the MessageSecurityMode to use.
        settings: A dictionary of additional settings to apply to the
            connection.
    """
    print(
        name,
        description,
        discoveryUrl,
        endpointUrl,
        securityPolicy,
        securityMode,
        settings,
    )


def callMethod(
    connectionName,  # type: String
    objectId,  # type: String
    methodId,  # type: String
    inputs,  # type: List[String]
):
    # type: (...) -> Tuple[Any, Any, Any]
    """Calls a method in an OPC UA Server.

    To make the most of this n, you'll need to be familiar with methods
    in the OPC UA Server.

    Args:
        connectionName: The name of the OPC UA connection to the Server
            that the method resides in.
        objectId: The NodeId of the Object Node the Method is a member
            of.
        methodId: The NodeId of the Method Node to call.
        inputs: A list of input values expected by the method.

    Returns:
        A tuple containing the following:
            0: Resulting StatusCode for the call
            1: A list of StatusCode objects corresponding to each
                input argument
            2: A list of output values

    """
    print(connectionName, objectId, methodId, inputs)
    return None, None, None


def removeConnection(name):
    # type: (String) -> bool
    """Removes an OPC UA Connection.

    Args:
        name: The name of the OPC UA connection to remove.

    Returns:
        A boolean value representing whether the function was able to
        remove the connection. Returns True if the connection was
        successfully removed, False if the connection was not removed.
    """
    print(name)
    return True



"""Date Functions.

The following functions give you access to test and modify dates.
"""

from __future__ import print_function

__all__ = [
    "addDays",
    "addHours",
    "addMillis",
    "addMinutes",
    "addMonths",
    "addSeconds",
    "addWeeks",
    "addYears",
    "daysBetween",
    "format",
    "fromMillis",
    "getAMorPM",
    "getDate",
    "getDayOfMonth",
    "getDayOfWeek",
    "getDayOfYear",
    "getHour12",
    "getHour24",
    "getMillis",
    "getMinute",
    "getMonth",
    "getQuarter",
    "getSecond",
    "getTimezone",
    "getTimezoneOffset",
    "getTimezoneRawOffset",
    "getYear",
    "hoursBetween",
    "isAfter",
    "isBefore",
    "isBetween",
    "isDaylightTime",
    "midnight",
    "millisBetween",
    "minutesBetween",
    "monthsBetween",
    "now",
    "parse",
    "secondsBetween",
    "setTime",
    "toMillis",
    "weeksBetween",
    "yearsBetween",
]

from datetime import datetime
from time import localtime, mktime

from typing import Optional

from java.lang import String
from java.util import Date, Locale


def _now():
    # type: () -> datetime
    """Get current datetime.

    Returns:
        Now.
    """
    return datetime.now()


def addDays(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of days to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addHours(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of hours to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addMillis(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of milliseconds to a given date and
    time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addMinutes(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of minutes to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addMonths(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of months to a given date and time.

    This function is unique since each month can have a variable number
    of days. For example, if the date passed in is March 31st, and we
    add one month, April does not have a 31st day, so the returned date
    will be the proper number of months rounded down to the closest
    available day, in this case April 30th.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addSeconds(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of seconds to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addWeeks(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of weeks to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def addYears(date, value):
    # type: (Date, int) -> Date
    """Add or subtract an amount of years to a given date and time.

    Args:
        date: The starting date.
        value: The number of units to add, or subtract if the value is
            negative.

    Returns:
        A new date object offset by the integer passed to the function.
    """
    print(date, value)
    return Date()


def daysBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole days between two dates.

    Daylight Saving Time changes are taken into account.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_1, date_2)
    return 1


def format(date, format="yyyy-MM-dd HH:mm:ss"):
    # type: (Date, String) -> unicode
    """Returns the given date as a string, formatted according to a
    pattern.

    The pattern is a format that is full of various placeholders that
    display different parts of the date. These are case-sensitive. These
    placeholders can be repeated for a different effect. For example, M
    will give you 1-12, MM will give you 01-12, MMM will give you
    Jan-Dec, MMMM will give you January-December.

    Note:
        The functions above are based off the Java class
        `SimpleDateFormat` internally. The functions will accept any
        valid format string for that class.

    Args:
        date: The date to format.
        format: A format string such as "yyyy-MM-dd HH:mm:ss". The
            format argument is optional. The default is
            "yyyy-MM-dd HH:mm:ss". Optional.

    Returns:
        A string representing the formatted datetime.
    """
    print(date, format)
    _date = _now()
    _format = format
    _format = _format.replace("a", "%p")
    _format = _format.replace("D", "%j")
    _format = _format.replace("dd", "{:02}".format(_date.day))
    _format = _format.replace("d", "{}".format(_date.day))
    _format = _format.replace("EEEE", "%A")
    _format = _format.replace("E", "%a")
    _format = _format.replace("F", "{}".format(_date.isocalendar()[2]))
    _format = _format.replace("G", "AD")
    _format = _format.replace("hh", "%I")
    _format = _format.replace("h", "%-I")
    _format = _format.replace("HH", "{:02}".format(_date.hour))
    _format = _format.replace("H", "%-H")
    _format = _format.replace("k", "%H")
    _format = _format.replace("K", "%-I")
    _format = _format.replace("MMMM", "%B")
    _format = _format.replace("MMM", "%b")
    _format = _format.replace("MM", "{:02}".format(_date.month))
    _format = _format.replace("M", "{}".format(_date.month))
    _format = _format.replace("mm", "%M")
    _format = _format.replace("m", "%-M")
    _millis = _date.microsecond % 1000
    _format = _format.replace("SSS", "{:03}".format(_millis))
    _format = _format.replace("SS", "{:02}".format(_millis))
    _format = _format.replace("S", "{:01}".format(_millis))
    _format = _format.replace("ss", "%S")
    _format = _format.replace("s", "%-S")
    _format = _format.replace("w", "{}".format(_date.isocalendar()[1] + 1))
    _format = _format.replace("u", "%w")
    _format = _format.replace(
        "W",
        "{}".format(_date.isocalendar()[1] - _date.replace(day=1).isocalendar()[1] + 1),
    )

    _format = _format.replace("XXX", "{:03.0f}:00".format(getTimezoneOffset()))
    _format = _format.replace("XX", "{:03.0f}00".format(getTimezoneOffset()))
    _format = _format.replace("X", "{:03.0f}".format(getTimezoneOffset()))
    _format = _format.replace("yyyy", "{}".format(_date.year))
    _format = _format.replace("yy", str(_date.year)[-2:])
    _format = _format.replace("Y", "%Y")
    _format = _format.replace("y", "%Y")
    _format = _format.replace("Z", "{:03.0f}00".format(getTimezoneOffset()))
    _format = _format.replace("z", "PDT")
    return unicode(_date.strftime(_format))


def fromMillis(millis):
    # type: (int) -> Date
    """Creates a date object given a millisecond value.

    Args:
        millis: The number of milliseconds elapsed since January 1,
            1970, 00:00:00 UTC (GMT).

    Returns:
        A new date object.
    """
    return Date(millis)


def getAMorPM(date):
    # type: (Date) -> int
    """Returns a 0 if the time is before noon, and a 1 if the time is
    after noon.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return 1 if _now().hour >= 12 else 0


def getDate(year, month, day):
    # type: (int, int, int) -> Date
    """Creates a new Date object given a year, month and a day.

    The time will be set to midnight of that day.

    Args:
        year: The year for the new date.
        month: The month of the new date. January is month 0.
        day: The day of the month for the new date. The first day
            of the month is day 1.

    Returns:
        A new date, set to midnight of that day.
    """
    print(year, month, day)
    return Date()


def getDayOfMonth(date):
    # type: (Date) -> int
    """Extracts the day of the month from a date.

    The first day of the month is day 1.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().day


def getDayOfWeek(date):
    # type: (Date) -> int
    """Extracts the day of the week from a date.

    Sunday is day 1, Saturday is day 7.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    _dow = [2, 3, 4, 5, 6, 7, 1]
    return _dow[_now().weekday()]


def getDayOfYear(date):
    # type: (Date) -> int
    """Extracts the day of the year from a date.

    The first day of the year is day 1.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().timetuple().tm_yday


def getHour12(date):
    # type: (Date) -> int
    """Extracts the hour from a date.

    Uses a 12 hour clock, so noon and midnight are returned as 0.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().hour - 12 if _now().hour >= 12 else _now().hour


def getHour24(date):
    # type: (Date) -> int
    """Extracts the hour from a date.

    Uses a 24 hour clock, so midnight is zero.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().hour


def getMillis(date):
    # type: (Date) -> int
    """Extracts the milliseconds from a date, ranging from 0-999.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().microsecond // 1000


def getMinute(date):
    # type: (Date) -> int
    """Extracts the minutes from a date, ranging from 0-59.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().minute


def getMonth(date):
    # type: (Date) -> int
    """Extracts the month from a date, where January is month 0.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().month


def getQuarter(date):
    # type: (Date) -> int
    """Extracts the quarter from a date, ranging from 1-4.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return ((_now().month - 1) // 3) + 1


def getSecond(date):
    # type: (Date) -> int
    """Extracts the seconds from a date, ranging from 0-59.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().second


def getTimezone():
    # type: () -> String
    """Returns the ID of the current timezone.

    Returns:
        A representation of the current timezone.
    """
    return "America/Tijuana"


def getTimezoneOffset(date=None):
    # type: (Optional[Date]) -> float
    """Returns the current timezone's offset versus UTC for a given
    instant, taking Daylight Saving Time into account.

    Args:
        date: The instant in time for which to calculate the offset.
            Uses now() if omitted. Optional.

    Returns:
        The timezone offset compared to UTC, in hours.
    """
    return -7.0 if isDaylightTime(date) else -8.0


def getTimezoneRawOffset():
    # type: () -> float
    """Returns the current timezone offset versus UTC, not taking
    Daylight Saving Time into account.

    Returns:
         The timezone offset.
    """
    return float(0)


def getYear(date):
    # type: (Date) -> int
    """Extracts the year from a date.

    Args:
        date: The date to use.

    Returns:
        An integer that is representative of the extracted value.
    """
    print(date)
    return _now().year


def hoursBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole hours between two dates.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_1, date_2)
    return 1


def isAfter(date_1, date_2):
    # type: (Date, Date) -> bool
    """Compares two dates to see if date_1 is after date_2.

    Args:
        date_1: The first date.
        date_2: The second date.

    Returns:
        True if date_1 is after date_2, False otherwise.
    """
    print(date_1, date_2)
    return True


def isBefore(date_1, date_2):
    # type: (Date, Date) -> bool
    """Compares to dates to see if date_1 is before date_2.

    Args:
        date_1: The first date.
        date_2: The second date.

    Returns:
        True if date_1 is before date_2, False otherwise.
    """
    print(date_1, date_2)
    return False


def isBetween(target_date, start_date, end_date):
    # type: (Date, Date, Date) -> bool
    """Compares two dates to see if a target date is between two other
    dates.

    Args:
        target_date: The date to compare.
        start_date: The start of a date range.
        end_date: The end of a date range. This date must be after the
            start date.

    Returns:
        True if target_date is >= start_date and target_date <=
        end_date, False otherwise.
    """
    print(target_date, start_date, end_date)
    return True


def isDaylightTime(date=None):
    # type: (Optional[Date]) -> bool
    """Checks to see if the current timezone is using Daylight Saving
    Time during the date specified.

    Args:
        date: The date you want to check if the current timezone is
            observing Daylight Saving Time. Uses now() if omitted.
            Optional.

    Returns:
        True if date is observing Daylight Saving Time in the
        current timezone, False otherwise.
    """
    print(date)
    _date = _now()
    time_tuple = (
        _date.year,
        _date.month,
        _date.day,
        _date.hour,
        _date.minute,
        _date.second,
        _date.weekday(),
        0,
        0,
    )
    stamp = mktime(time_tuple)
    time_tuple = localtime(stamp)
    return time_tuple.tm_isdst > 0


def midnight(date):
    # type: (Date) -> Date
    """Returns a copy of a date with the hour, minute, second, and
    millisecond fields set to zero.

    Args:
        date: The starting date.

    Returns:
        A new date, set to midnight of the day provided.
    """
    print(date)
    return Date()


def millisBetween(date_1, date_2):
    # type: (Date, Date) -> long
    """Calculates the number of whole milliseconds between two dates.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_2, date_1)
    return long(1)


def minutesBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole minutes between two dates.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_2, date_1)
    return 1


def monthsBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole months between two dates.

    Daylight Saving Time changes are taken into account.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_1, date_2)
    return 0


def now():
    # type: () -> Date
    """Returns a java.util.Date object that represents the current time
    according to the local system clock.

    Returns:
        A new date, set to the current date and time.
    """
    return Date()


def parse(dateString, formatString="yyyy-MM-dd HH:mm:ss", locale=Locale.ENGLISH):
    # type: (String, Optional[String], Optional[Locale]) -> Date
    """Attempts to parse a string and create a Date.

    Causes ParseException if the date dateString parameter is in an
    unrecognized format.

    Args:
        dateString: The string to parse into a date.
        formatString: Format string used by the parser. Default is
            "yyyy-MM-dd HH:mm:ss". Optional.
        locale: Locale used for parsing. Can be the locale name
            such as 'fr', or the Java Locale such as 'Locale.FRENCH'.
            Default is 'Locale.ENGLISH'. Optional.

    Returns:
        The parsed date.
    """
    print(dateString, formatString, locale)
    return Date()


def secondsBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole seconds between two dates.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_2, date_1)
    return 0


def setTime(date, hour, minute, second):
    # type: (Date, int, int, int) -> Date
    """Takes in a date, and returns a copy of it with the time fields
    set as specified.

    Args:
        date: The starting date.
        hour: The hours (0-23) to set.
        minute: The minutes (0-59) to set.
        second: The seconds (0-59) to set.

    Returns:
        A new date, set to the appropriate time.
    """
    print(date, hour, minute, second)
    return Date()


def toMillis(date):
    # type: (Date) -> long
    """Converts a Date object to its millisecond value elapsed since
    January 1, 1970, 00:00:00 UTC (GMT).

    Args:
        date: The date object to convert.

    Returns:
        An 8-byte integer representing the number of millisecond elapsed
        since January 1, 1970, 00:00:00 UTC (GMT).
    """
    print(date)
    _date = _now()
    millis = mktime(_date.timetuple()) * 1000 + _date.microsecond // 1000
    return long(millis)


def weeksBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole weeks between two dates.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_1, date_2)
    return 0


def yearsBetween(date_1, date_2):
    # type: (Date, Date) -> int
    """Calculates the number of whole years between two dates.

    Daylight Saving Time changes are taken into account.

    Args:
        date_1: The first date to use.
        date_2: The second date to use.

    Returns:
        An integer that is representative of the difference between two
        dates.
    """
    print(date_1, date_2)
    return 0



"""Database Functions.

The following functions give you access to view and modify data in the
database.
"""

from __future__ import print_function

__all__ = [
    "ARRAY",
    "BIGINT",
    "BINARY",
    "BIT",
    "BLOB",
    "BOOLEAN",
    "CHAR",
    "CLOB",
    "DATALINK",
    "DATE",
    "DECIMAL",
    "DISTINCT",
    "DOUBLE",
    "FLOAT",
    "INTEGER",
    "JAVA_OBJECT",
    "LONGNVARCHAR",
    "LONGVARBINARY",
    "LONGVARCHAR",
    "NCHAR",
    "NCLOB",
    "NULL",
    "NUMERIC",
    "NVARCHAR",
    "ORACLE_CURSOR",
    "OTHER",
    "READ_COMMITTED",
    "READ_UNCOMMITTED",
    "REAL",
    "REF",
    "REPEATABLE_READ",
    "ROWID",
    "SERIALIZABLE",
    "SMALLINT",
    "SQLXML",
    "STRUCT",
    "TIME",
    "TIMESTAMP",
    "TINYINT",
    "VARBINARY",
    "VARCHAR",
    "addDatasource",
    "beginNamedQueryTransaction",
    "beginTransaction",
    "clearAllNamedQueryCaches",
    "clearNamedQueryCache",
    "closeTransaction",
    "commitTransaction",
    "createSProcCall",
    "dateFormat",
    "execSProcCall",
    "getConnectionInfo",
    "getConnections",
    "refresh",
    "removeDatasource",
    "rollbackTransaction",
    "runNamedQuery",
    "runPrepQuery",
    "runPrepUpdate",
    "runQuery",
    "runSFNamedQuery",
    "runSFPrepUpdate",
    "runSFUpdateQuery",
    "runScalarPrepQuery",
    "runScalarQuery",
    "runUpdateQuery",
    "setDatasourceConnectURL",
    "setDatasourceEnabled",
    "setDatasourceMaxConnections",
]

from typing import Any, List, Optional

from com.inductiveautomation.ignition.common import BasicDataset
from com.inductiveautomation.ignition.common.script.builtin import (
    DatasetUtilities,
    SProcCall,
)
from java.lang import String
from java.util import Date
from javax.swing import JComponent

# Type codes
# These are codes defined by the JDBC specification.
BIT = -7
REAL = 7
LONGVARCHAR = -1
LONGVARBINARY = -4
TINYINT = -6
DOUBLE = 8
DATE = 91
NULL = 0
SMALLINT = 5
NUMERIC = 2
TIME = 92
ROWID = -8
INTEGER = 4
DECIMAL = 3
TIMESTAMP = 93
CLOB = 2005
BIGINT = -5
CHAR = 1
BINARY = -2
NCLOB = 2011
FLOAT = 6
VARCHAR = 12
VARBINARY = -3
BLOB = 2004
NCHAR = -15
NVARCHAR = -9
LONGNVARCHAR = -16
BOOLEAN = 16
# The following type code constants are available for other uses, but
# are not supported by createSProcCall:
ORACLE_CURSOR = -10
DISTINCT = 2001
STRUCT = 2002
REF = 2006
JAVA_OBJECT = 2000
SQLXML = 2009
ARRAY = 2003
DATALINK = 70
OTHER = 1111

# Isolation levels.
READ_COMMITTED = 2
READ_UNCOMMITTED = 1
REPEATABLE_READ = 4
SERIALIZABLE = 8


def addDatasource(
    jdbcDriver,  # type: String
    name,  # type: String
    description="",  # type: String
    connectUrl=None,  # type: Optional[String]
    username=None,  # type: Optional[String]
    password=None,  # type: Optional[String]
    props=None,  # type: Optional[String]
    validationQuery=None,  # type: Optional[String]
    maxConnections=8,  # type: int
):
    # type: (...) -> None
    """Adds a new database connection in Ignition.

    Args:
        jdbcDriver: The name of the JDBC driver configuration to use.
            Available options are based off the JDBC driver
            configurations on the Gateway.
        name: The datasource name.
        description: Description of the datasource. Optional.
        connectUrl: Default is the connect URL for JDBC driver.
            Optional.
        username: Username to login to the datasource with. Optional.
        password: Password for the login. Optional.
        props: The extra connection parameters. Optional.
        validationQuery: Default is the validation query for the JDBC
            driver. Optional.
        maxConnections: Default is 8. Optional.
    """
    print(
        jdbcDriver,
        name,
        description,
        connectUrl,
        username,
        password,
        props,
        validationQuery,
        maxConnections,
    )


def beginNamedQueryTransaction(*args):
    # type: (*Any) -> String
    """Begins a new database transaction using Named Queries.

    Database transactions are used to execute multiple queries in an
    atomic fashion. After executing queries, you must either commit the
    transaction to have your changes take effect, or rollback the
    transaction which will make all operations since the last commit not
    take place. The transaction is given a new unique string code, which
    is then returned. You can then use this code as the tx argument for
    other system.db.* function calls to execute various types of queries
    using this transaction.

    An open transaction consumes one database connection until it is
    closed. Because leaving connections open indefinitely would exhaust
    the connection pool, each transaction is given a timeout. Each time
    the transaction is used, the timeout timer is reset. For example, if
    you make a transaction with a timeout of one minute, you must
    complete that transaction within a minute. If a transaction is
    detected to have timed out, it will be automatically closed and its
    transaction id will no longer be valid.

    When calling from Vision use:
    system.db.beginNamedQueryTransaction([database], [isolationLevel],
                                         [timeout])

    When calling from Perspective or Gateway scope use:
    system.db.beginNamedQueryTransaction(project, database,
                                        [isolationLevel], [timeout])

    Args:
        *args: Variable length argument list.

    Returns:
         The new transaction ID. You'll use this ID as the "tx" argument
         for all other calls to have them execute against this
         transaction.
    """
    print(args)
    return "transaction_id"


def beginTransaction(database="", isolationLevel=None, timeout=None):
    # type: (Optional[String], Optional[int], Optional[int]) -> String
    """Begins a new database transaction for using run* and runPrep*
    queries.

    Database transactions are used to execute multiple queries in an
    atomic fashion. After executing queries, you must either commit the
    transaction to have your changes take effect, or rollback the
    transaction which will make all operations since the last commit not
    take place. The transaction is given a new unique string code, which
    is then returned. You can then use this code as the tx argument for
    other system.db.* function calls to execute various types of queries
    using this transaction.

    An open transaction consumes one database connection until it is
    closed. Because leaving connections open indefinitely would exhaust
    the connection pool, each transaction is given a timeout. Each time
    the transaction is used, the timeout timer is reset. For example, if
    you make a transaction with a timeout of one minute, you must use
    that transaction at least once a minute. If a transaction is
    detected to have timed out, it will be automatically closed and its
    transaction id will no longer be valid.

    Args:
        database: The name of the database connection to create a
            transaction in. Use "" for the project's default connection.
            Optional.
        isolationLevel: The transaction isolation level to use. Use one
            of the four constants: system.db.READ_COMMITTED,
            system.db.READ_UNCOMMITTED, system.db.REPEATABLE_READ, or
            system.db.SERIALIZABLE. Optional.
        timeout: The amount of time, in milliseconds, that this
            connection is allowed to remain open without being used.
            Timeout counter is reset any time a query or call is
            executed against the transaction, or when committed or
            rolled-back. Optional.

    Returns:
        The new transaction ID. You'll use this ID as the "tx" argument
        for all other calls to have them execute against this
        transaction.
    """
    print(database, isolationLevel, timeout)
    return "transaction_id"


def clearAllNamedQueryCaches(project=None):
    # type: (Optional[String]) -> None
    """This clears the caches of all Named Queries in a project.

    If called from the Shared Scope (i.e., Tag Event Scripts, Alarm
    Pipelines, etc.) then the name of the project must be passed as a
    parameter.

    Args:
        project: The project that contains the Named Query whose cache
            needs to be cleared. Optional.
    """
    print(project)


def clearNamedQueryCache(*args):
    # type: (*String) -> None
    """This clears the cache of a Named Query.

    If called from the Shared Scope (i.e., Tag Event Scripts, Alarm
    Pipelines, etc.) then the name of the project must be passed as a
    parameter.

    When calling from the Project Scope use:
    system.db.clearNamedQueryCache(path)

    When calling from the Shared Scope use:
    system.db.clearNamedQueryCache(project, path)

    Args:
        *args: Variable length argument list.
    """
    print(args)


def closeTransaction(tx):
    # type: (String) -> None
    """Closes the transaction with the given ID.

    You must commit or rollback the transaction before you close it.
    Closing the transaction will return its database connection to the
    pool. The transaction ID will no longer be valid.

    Args:
        tx: The transaction ID.
    """
    print(tx)


def commitTransaction(tx):
    # type: (String) -> None
    """Performs a commit for the given transaction.

    This will make all statements executed against the transaction since
    its beginning or since the last commit or rollback take effect in
    the database. Until you commit a transaction, any changes that the
    transaction makes will not be visible to other connections.

    Note:
        If you are done with the transaction, you must close it after
        you commit it.

    Args:
        tx: The transaction ID.
    """
    print(tx)


def createSProcCall(
    procedureName,  # type: String
    database="",  # type: String
    tx=None,  # type: Optional[String]
    skipAudit=False,  # type: bool
):
    # type: (...) -> SProcCall
    """Creates an SProcCall object, which is a stored procedure call
    context.

    Args:
        procedureName: The name of the stored procedure to call.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        tx: A transaction identifier. If omitted, the call will be
            executed in its own transaction. Optional.
        skipAudit: A flag which, if set to True, will cause the
            procedure call to skip the audit system. Useful for some
            queries that have fields which won't fit into the audit log.
            Optional.

    Returns:
        A stored procedure call context, which can be configured and
        then used as the argument to system.db.execSProcCall.
    """
    print(procedureName, database, tx, skipAudit)
    return SProcCall()


def dateFormat(date, formatPattern):
    # type: (Date, String) -> String
    """This function is used to format Dates nicely as strings.

    It uses a format string to guide its formatting behavior.

    Args:
        date: The Date object that you'd like to format.
        formatPattern: A format pattern string to apply.

    Returns:
        The date as a string formatted according to the format pattern.
    """
    print(date, formatPattern)
    return ""


def execSProcCall(callContext):
    # type: (SProcCall) -> None
    """Executes a stored procedure call.

    The one parameter to this function is an SProcCall - a stored
    procedure call context.

    Args:
        callContext: A stored procedure call context, with any input,
            output, and/or return value parameters correctly configured.
            Use system.db.createSProcCall to create a call context.
    """
    print(callContext)


def getConnectionInfo(name=""):
    # type: (Optional[String]) -> BasicDataset
    """Returns a dataset of information about a single database
    connection, as specified by the name argument, or about the current
    project's default database connection.

    Note:
        The database connection used when called from the Gateway scope
        is the connection configured on the Gateway scripting project.

    Args:
        name: The name of the database connection to find information
            about. Optional.

    Returns:
        A dataset containing information about the named database
        connection or about the current project's default database
        connection, or an empty dataset if the connection wasn't found.
    """
    print(name)
    return BasicDataset()


def getConnections():
    # type: () -> BasicDataset
    """Returns a dataset of information about each configured database
    connection.

    Each row represents a single connection.

    Returns:
        A dataset, where each row represents a database connection.
    """
    return BasicDataset()


def refresh(component, propertyName):
    # type: (JComponent, String) -> bool
    """This function will cause a Vision component binding to execute
    immediately.

    This is most often used for bindings that are set to Polling - Off.
    In this way, you cause a binding to execute on demand, when you know
    that the results of its query will return a new result. To use it,
    you simply specify the component and name of the property on whose
    binding you'd like to refresh.

    Even though the function includes "db" in the name, the function can
    update all types of Vision component bindings, including Property
    and Expression bindings.

    Note:
        This function will only work within the Vision module. To
        manually execute bindings in Perspective, use the refreshBinding
        component method.

    Args:
        component: The component whose property you want to refresh.
        propertyName: The name of the property that has a binding that
            needs to be refreshed.

    Returns:
        True if the property was found and refreshed successfully.
    """
    print(component, propertyName)
    return True


def removeDatasource(name):
    # type: (String) -> None
    """Removes a database connection from Ignition.

    Args:
        name: The name of the database connection in Ignition.
    """
    print(name)


def rollbackTransaction(tx):
    # type: (String) -> None
    """Performs a rollback on the given connection.

    This will make all statements executed against this transaction
    since its beginning or since the last commit or rollback undone.

    Note:
        If you are done with the transaction, you must also close it
        after you do a rollback on it.

    Args:
        tx: The transaction ID.
    """
    print(tx)


def runNamedQuery(*args):
    # type: (*Any) -> Any
    """Runs a named query and returns the results.

    Note that the number of parameters in the function is determined by
    scope. Both versions of the function are listed below.

    When calling from the Project Scope use:
    system.db.runNamedQuery(path, parameters, [tx], [getKey])

    When calling from the Gateway Scope use:
    system.db.runNamedQuery(project, path, parameters, [tx], [getKey])

    Args:
        *args: Variable length argument list.

    Returns:
        The results of the query. The exact object returned depends on
        the Query Type property of the Named Query: typically either a
        dataset when set to Query, an integer representing the number of
        rows affected when set to Update Query, or an object matching
        the datatype of the value returned by a Scalar Query.
    """
    print(args)


def runPrepQuery(
    query,  # type: String
    args,  # type: List[Any]
    database="",  # type: String
    tx=None,  # type: Optional[String]
):
    # type: (...) -> DatasetUtilities.PyDataSet
    """Runs a prepared statement against the database, returning the
    results in a PyDataSet.

    Prepared statements differ from regular queries in that they can use
    a special placeholder, the question-mark character (?) in the query
    where any dynamic arguments would go, and then use an array of
    values to provide real information for those arguments. Make sure
    that the length of your argument array matches the number of
    question-mark placeholders in your query.

    This call should be used for SELECT queries. This is a useful
    alternative to system.db.runQuery because it allows values in the
    WHERE clause, JOIN clause, and other clauses to be specified without
    having to turn those values into strings. This is safer because it
    protects against a problem known as a SQL injection attack, where a
    user can input data that affects the query's semantics.

    Note:
        The "?" placeholder refers to variables of the query statement
        that help the statement return the correct information. The "?"
        placeholder cannot reference column names, table names, or the
        underlying syntax of the query. This is because the SQL standard
        for handling the "?" placeholder excludes these items.

    Args:
        query: A query (typically a SELECT) to run as a prepared
            statement with placeholders (?) denoting where the arguments
            go.
        args: A list of arguments. Will be used in order to match each
            placeholder (?) found in the query.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used.
        tx: A transaction identifier. If omitted, the query will be
            executed in its own transaction. Optional.

    Returns:
        The results of the query as a PyDataSet.
    """
    print(query, args, database, tx)
    return DatasetUtilities.PyDataSet()


def runPrepUpdate(
    query,  # type: String
    args,  # type: List[Any]
    database="",  # type: String
    tx=None,  # type: Optional[String]
    getKey=False,  # type: bool
    skipAudit=True,  # type: bool
):
    # type: (...) -> int
    """Runs a prepared statement against the database, returning the
    number of rows that were affected.

    Prepared statements differ from regular queries in that they can use
    a special placeholder, the question-mark character (?) in the query
    where any dynamic arguments would go, and then use an array of
    values to provide real information for those arguments. Make sure
    that the length of your argument array matches the number of
    question-mark placeholders in your query. This call should be used
    for UPDATE, INSERT, and DELETE queries.

    Note:
        The "?" placeholder refers to variables of the query statement
        that help the statement return the correct information. The "?"
        placeholder cannot reference column names, table names, or the
        underlying syntax of the query. This is because the SQL standard
        for handling the "?" placeholder excludes these items.

    Args:
        query: A query (typically an UPDATE, INSERT, or DELETE) to run
            as a prepared statement with placeholders (?) denoting where
            the arguments go.
        args: A list of arguments. Will be used in order to match each
            placeholder (?) found in the query.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        tx: A transaction identifier. If omitted, the update will be
            executed in its own transaction. Optional.
        getKey: A flag indicating whether or not the result should be
            the number of rows returned (getKey=0) or the newly
            generated key value that was created as a result of the
            update (getKey=1). Not all databases support automatic
            retrieval of generated keys. Optional.
        skipAudit: A flag which, if set to True, will cause the prep
            update to skip the audit system. Useful for some queries
            that have fields which won't fit into the audit log.
            Optional.

    Returns:
        The number of rows affected by the query, or the key value that
        was generated, depending on the value of the getKey flag.
    """
    print(query, args, database, tx, getKey, skipAudit)
    return 1


def runQuery(
    query,  # type: String
    database="",  # type: String
    tx=None,  # type: Optional[String]
):
    # type: (...) -> DatasetUtilities.PyDataSet
    """Runs a SQL query, usually a SELECT query, against a database,
    returning the results as a dataset.

    If no database is specified, or the database is the empty-string "",
    then the current project's default database connection will be used.
    The results are returned as a PyDataSet, which is a wrapper around
    the standard dataset that is convenient for scripting.

    Args:
        query: A SQL query, usually a SELECT query, to run.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used.
        tx: A transaction identifier. If omitted, the query will be
            executed in its own transaction.

    Returns:
        The results of the query as a PyDataSet.
    """
    print(query, database, tx)
    return DatasetUtilities.PyDataSet()


def runSFNamedQuery(*args):
    # type: (*Any) -> bool
    """Runs a named query that goes through the Store and Forward
    system.

    Note that the number of parameters in the function is determined by
    scope.

    When calling from the Project Scope use:
    system.db.runSFNamedQuery(path, params)

    When calling from the Gateway Scope use:
    system.db.runSFNamedQuery(project, path, params)

    Args:
        *args : Variable length argument list.

    Returns:
        True if successfully sent to the Store and Forward system.
    """
    print(args)
    return True


def runSFPrepUpdate(query, args, datasources):
    # type: (String, List[Any], List[String]) -> bool
    """Runs a prepared statement query through the store and forward
    system and to multiple datasources at the same time.

    Prepared statements differ from regular queries in that they can use
    a special placeholder, the question-mark character (?) in the query
    where any dynamic arguments would go, and then use an array of
    values to provide real information for those arguments. Make sure
    that the length of your argument array matches the number of
    question-mark placeholders in your query. This call should be used
    for UPDATE, INSERT, and DELETE queries.

    Args:
        query: A query (typically an UPDATE, INSERT, or DELETE) to run
            as a prepared statement, with placeholders (?) denoting
            where the arguments go.
        args: A list of arguments. Will be used in order to match each
            placeholder (?) found in the query.
        datasources: List of datasources to run the query through.

    Returns:
        True if successfully sent to Store and Forward system.
    """
    print(query, args, datasources)
    return True


def runSFUpdateQuery(query, datasources):
    # type: (String, List[String]) -> bool
    """Runs an query through the store and forward system and to
    multiple datasources at the same time.

    Args:
        query: A query (typically an UPDATE, INSERT, or DELETE) to run.
        datasources: List of datasources to run the query through.

    Returns:
        True if successful and False if not.
    """
    print(query, datasources)
    return True


def runScalarPrepQuery(
    query,  # type: String
    args,  # type: List[Any]
    database="",  # type: String
    tx=None,  # type: Optional[String]
):
    # type: (...) -> Any
    """Runs a prepared statement against a database connection just like
    the runPrepQuery function, but only returns the value from the first
    row and column.

    If no results are returned from the query, the special value None is
    returned.

    Args:
        query: A SQL query (typically a SELECT) to run as a prepared
            statement with placeholders (?) denoting where the arguments
            go, that should be designed to return one row and one
            column.
        args: A list of arguments. Will be used in order to match each
            placeholder (?) found in the query.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        tx: A transaction identifier. If omitted, the query will be
            executed in its own transaction. Optional.

    Returns:
         The value from the first row and first column of the results.
         Returns None if no rows were returned.
    """
    print(query, args, database, tx)


def runScalarQuery(query, database="", tx=None):
    # type: (String, Optional[String], Optional[String]) -> Any
    """Runs a query against a database connection just like the runQuery
    function, but only returns the value from the first row and column.

    If no results are returned from the query, the special value None is
    returned.

    Args:
        query: A SQL query that should be designed to return one row and
            one column.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used. Optional.
        tx: A transaction identifier. If omitted, the query will be
            executed in its own transaction. Optional.

    Returns:
         The value from the first row and first column of the results.
         Returns None if no rows were returned.
    """
    print(query, database, tx)


def runUpdateQuery(
    query,  # type: String
    database="",  # type: String
    tx=None,  # type: Optional[String]
    getKey=False,  # type: bool
    skipAudit=True,  # type: bool
):
    # type: (...) -> int
    """Runs a query against a database connection, returning the number
    of rows affected.

    Typically this is an UPDATE, INSERT, or DELETE query. If no database
    is specified, or the database is the empty-string "", then the
    current project's default database connection will be used.

    Note:
        You may want to use the runPrepUpdate query if your query is
        constructed with user input (to avoid the user's input from
        breaking your syntax) or if you need to insert binary or BLOB
        data.

    Args:
        query: A SQL query, usually an INSERT, UPDATE, or DELETE query,
            to run.
        database: The name of the database connection to execute
            against. If omitted or "", the project's default database
            connection will be used.
        tx: A transaction identifier. If omitted, the update will be
            executed in its own transaction.
        getKey: A flag indicating whether or not the result should be
            the number of rows returned (getKey=0) or the newly
            generated key value that was created as a result of the
            update (getKey=1). Not all databases support automatic
            retrieval of generated keys.
        skipAudit: A flag which, if set to True, will cause the update
            query to skip the audit system. Useful for some queries that
            have fields which won't fit into the audit log.

    Returns:
        The number of rows affected by the query, or the key value that
        was generated, depending on the value of the getKey flag.
    """
    print(query, database, tx, getKey, skipAudit)
    return 1


def setDatasourceConnectURL(name, connectUrl):
    # type: (String, String) -> None
    """Changes the connect URL for a given database connection.

    Args:
        name: The name of the database connection in Ignition.
        connectUrl: The new connect URL.
    """
    print(name, connectUrl)


def setDatasourceEnabled(name, enabled):
    # type: (String, bool) -> None
    """Enables/disables a given database connection.

    Args:
        name: The name of the database connection in Ignition.
        enabled: Specifies whether the database connection will be set
            to enabled or disabled state.
    """
    print(name, enabled)


def setDatasourceMaxConnections(name, maxConnections):
    # type: (String, int) -> None
    """Sets the Max Active and Max Idle parameters of a given database
    connection.

    Ags:
        name: The name of the database connection in Ignition.
        maxConnections: The new value for Max Active and Max Idle.
    """
    print(name, maxConnections)
